import random

from aardtools.sortexternal import SortExternal

def test_with_ints():
    s = SortExternal()
    for i in range(100000):
        line = "%08i" % random.randint(0, 99999999)
        s.put(line)
    s.sort()
    prev = 0
    for line in s:
        val = int(line)
        assert val >= prev
    s.cleanup()



#coding: utf8
# This file is part of Aard Dictionary Tools <http://aarddict.org>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License <http://www.gnu.org/licenses/gpl-3.0.txt>
# for more details.
#
# Copyright (C) 2008-2009  Igor Tkach

import functools
import logging
from collections import defaultdict

EXCLUDE_CLASSES = set(('navbox', 'collapsible', 'autocollapse', 'plainlinksneverexpand', 'navbar'))

def convert(obj):
    w = MWAardWriter()
    text, tags = w.txt(obj)
    return text, tags

def newline(func):
    def f(*args, **kwargs):
        txt, tags = func(*args, **kwargs)
        if txt and not txt.endswith(u'\n'):
            txt += u'\n'
        return txt, tags
    f.__name__ = func.__name__
    f.__doc__ = func.__doc__
    return f

class Cell(object):

    def __init__(self, text, tags=None, colspan=1, rowspan=1):
        self.text = text.rstrip()
        self.colspan = colspan
        self.rowspan = rowspan
        self.tags = [] if tags is None else tags

    def __str__(self):
        return 'Cell("""%s""", %s, %s, %s)' % (self.text.encode('utf8'),
                                               self.tags, self.colspan,
                                               self.rowspan)

    def __repr__(self):
        return 'Cell("""%s""", %s, %s, %s)' % (self.text.encode('utf8'),
                                               self.tags, self.colspan,
                                               self.rowspan)


class Row(list):
    def __init__(self, attributes=None):
        self.attributes = attributes

class MWAardWriter(object):

    def __init__(self):
        self.refgroups = defaultdict(list)
        self.errors = []
        self.languagelinks = []
        self.categorylinks = []
        self.current_list_number = 0
        self.current_tables = []

    def _Text(self, obj):
        return obj.caption, []

    @newline
    def _ItemList(self, obj):
        self.current_list_number = 0
        return self.process_children(obj)

    @newline
    def _Item(self, obj):
        txt = u''
        if (obj.parent.numbered):
            self.current_list_number += 1
            txt += u'%d. ' % self.current_list_number
        else:
            txt += u'\u2022 '
        return self.process_children(obj, txt)

    @newline
    def _Paragraph(self, obj):
        txt, tags = self.process_children(obj)
        tags.append(maketag(u'p', txt))
        return txt, tags

    @newline
    def _Section(self, obj):
        level = 2 + obj.getLevel() # starting with h2
        h = u'h%d' % level
        txt, tags = self.txt(obj.children[0])
        tags.append(maketag(h, txt))
        txt += u'\n'
        obj.children = obj.children[1:]
        return self.process_children(obj, txt, tags)

    @newline
    def _Chapter(self, obj):
        txt = obj.caption
        tags = [maketag(u'h1', txt)]
        return txt, tags

    @newline
    def _Caption(self, obj):
        txt = obj.caption
        tags = [maketag(u'strong', txt)]
        return txt, tags

    def _CategoryLink(self, obj):
        self.categorylinks.append(obj.target)
        return u'', []

    def _LangLink(self, obj):
        self.languagelinks.append((obj.namespace, obj.target))
        return u'', []

    def _ArticleLink(self, obj):
        txt, tags = self.process_children(obj)
        if not txt:
            txt = obj.caption
        if not txt:
            txt = obj.target
        tags.append(maketag(u'a', txt, {u'href':obj.target}))
        return txt, tags

    def _NamespaceLink(self, obj):
        txt, tags = self.process_children(obj)
        target = obj.target
        if target and ':' in target:
            target = target[target.index(':')+1:]
        if not txt:
            txt = obj.caption
        if not txt:
            txt = target
        tags.append(maketag(u'a', txt, {u'href': target}))
        return txt, tags

    _InterwikiLink = _ArticleLink

    def _NamedURL(self, obj):
        txt, tags = self.process_children(obj)
        if not txt:
            txt = obj.caption
        tags.append(maketag(u'a', txt, {u'href':obj.caption}))
        return txt, tags

    def _URL(self, obj):
        txt, tags = self.process_children(obj)
        if not txt:
            txt = obj.caption
        tags.append(maketag(u'a', txt, {u'href':obj.caption}))
        return txt, tags

    def _Style(self, obj):
        txt, tags = self.process_children(obj)
        if obj.caption == "''":
            tags.append(maketag(u'i', txt))
        elif obj.caption == "'''":
            tags.append(maketag(u'b', txt))
        elif obj.caption == ";":
            tags.append(maketag(u'tt', txt))
        return txt, tags

    def _Teletyped(self, obj):
        txt, tags = self.process_children(obj)
        tags.append(maketag(u'tt', txt))
        return txt, tags

    def _TagNode(self, obj):
        txt, tags = self.process_children(obj)
        tagname = obj.caption
        tags.append(maketag(tagname, txt))
        return txt, tags

    def _Node(self, obj):
        txt, tags = self.process_children(obj)
        tagname = obj._tag if hasattr(obj, '_tag') else obj.caption
        if tagname:
            tags.append(maketag(tagname, txt))
        return txt, tags

    def _Math(self, obj):
        txt, tags = self.process_children(obj)
        return obj.caption + txt, tags

    def _ImageLink(self, obj):
        return '', []

    def _BreakingReturn(self, obj):
        return u'\n', []

    def _Generic(self, obj):
        classes = obj.attributes.get('class', '').split()
        if any((cl in EXCLUDE_CLASSES
                for cl in classes)):
            return '', []
        txt, tags = self.process_children(obj)
        tagname = obj._tag
        tags.append(maketag(tagname, txt))
        return txt, tags

    _Emphasized = _Generic
    _Strong = _Generic
    _Small = _Generic
    _Big = _Generic
    _Cite = _Generic
    _Sub = _Generic
    _Sup = _Generic
    _Font = _Generic
    _DefinitionDescription = _Generic
    _Div = newline(_Generic)

    def add_ref(self, obj):
        name = obj.attributes.get(u'name', '')
        group = obj.attributes.get(u'group', '')

        references = self.refgroups[group]

        refid = None

        if name:
            existing = [r for r in references
                        if name == r.attributes.get(u'name', '')]
            if existing:
                refid = references.index(existing[0])

        if refid is None:
            references.append(obj)
            refid = len(references)
        return refid, name, group

    def _Reference(self, obj):
        refid, name, group = self.add_ref(obj)
        refidstr = unicode(refid)
        txt = u'%s %s' % (group, refidstr)
        txt = u'[%s]' % txt.strip()
        return txt, [maketag(u'ref', txt, {u'id': refidstr,
                                           u'group': group})]

    @newline
    def _ReferenceList(self, obj):
        group = obj.attributes.get(u'group', '')
        tags = []
        txt = u''
        for i, refobj in enumerate(self.refgroups[group]):
            start = len(txt)
            refid = unicode(i+1)
            refidtxt = u'[%s]' % refid
            txt += refidtxt + u' '
            tags.append(maketag(u'b', refidtxt, start=start))
            txt, tags = self.process_children(refobj, txt, tags)
            tags.append(maketag(u'note', txt,  {u'id': refid,
                                                u'group': group},
                                                start=start))
            txt += u'\n'
        del self.refgroups[group]
        return txt, tags

    def _Article(self, a):
        # add article name as first section heading
        txt = a.caption
        logging.debug('Article %s', txt.encode('utf8'))
        tags = [maketag(u'h1', txt)]
        txt += u'\n'
        return self.process_children(a, txt, tags)

    def _Cell(self, obj):
        current_table, current_row = self.current_tables[-1]
        if current_row is None:
            logging.warn("Can't add cell outside of row")
        else:            
            txt, tags = self.process_children(obj)
            #txt = txt.replace('\n', ' ')
            txt = txt.replace('\t', ' ')
            colspan = obj.colspan
            rowspan = obj.rowspan
            #tags.append(maketag(u'td', obj.attributes))
            current_row.append(Cell(txt, tags, colspan, rowspan))
        return '', []

    def _Row(self, obj):
        current_table, current_row = self.current_tables[-1]
        if current_row is not None:
            logging.error('Processing row is already in progress')
        else:
            self.current_tables[-1] = (current_table, Row())
            self.process_children(obj)
            current_table, current_row = self.current_tables[-1]
            current_row.attributes = obj.attributes
            current_table.append(current_row)
            self.current_tables[-1] = (current_table, None)
        return '', []

    @newline
    def _Table(self, obj):
        tableclasses = obj.attributes.get('class', '').split()
        if any((tableclass in EXCLUDE_CLASSES
                for tableclass in tableclasses)):
            return '', []

        self.current_tables.append(([], None))
        if obj.caption:
            logging.debug('Table %s', obj.caption.encode('utf8'))
        tabletext, tags = self.process_children(obj)
        current_table, current_row = self.current_tables.pop()
        if current_table:
            txt = u' '
            tabletext, tags, tabs = self.maketable(current_table, tabletext, tags)
            if not tabs:
                tags.append(maketag('blockquote', tabletext))
                tabletext += u'\n'
                return tabletext, tags
            else:
                tags = [maketag('tbl', txt, {u'text': tabletext,
                                             u'tags': tags,
                                             u'tabs': tabs,
                                             })]
                return txt, tags
        else:
            caption = obj.caption if obj.caption else u''
            logging.warn('Table %s has no data, skipping',
                         caption.encode('utf8'))
            return '', []

    def maketable(self, data, tabletext=u'', tabletags=None):
        newdata = []
        rowspanmap = defaultdict(int)
        for i, row in enumerate(data):
            newrow = []
            j = 0
            for cell in row:
                while rowspanmap[j] > 0:
                    rowspanmap[j] = rowspanmap[j] - 1
                    j += 1
                    newrow.append(Cell(''))
                rowspan = cell.rowspan
                for k in range(j, j+cell.colspan):
                    rowspanmap[k] = rowspan - 1
                cell.rowspan = 1
                newrow.append(cell)
                j += cell.colspan
            newdata.append(newrow)

        tabcount = max([sum(cell.colspan for cell in row) for row in newdata])

        for i, row in enumerate(newdata):
            span = sum(cell.colspan for cell in row)
            if (span < tabcount):
                logging.debug('Bad row. Span %d expected, but only %d found',
                             tabcount, span)
            for p in range(tabcount - span):
                row.append(Cell(''))


        newdata2 = []
        for i, row in enumerate(newdata):
            linecounts = [len(cell.text.splitlines()) for cell in row]
            count = max(linecounts) if linecounts else 0
            if count > 1:
                newrows = [[] for k in range(count)]
                for cell in row:
                    lines = cell.text.splitlines()
                    while len(lines) < count:
                        lines.append('')

                    linelens = [len(line) for line in lines]
                    cutpoints = []
                    runningsum = 0
                    for linelen in linelens:
                        runningsum += (linelen + 1) # add one to take into account newline char
                        cutpoints.append(runningsum)

                    linetags=[[] for k in range(count)]
                    for tag in cell.tags:
                        start = tag[1]
                        end = tag[2]

                        prevcutpoint = 0
                        for c, cutpoint in enumerate(cutpoints):
                            if prevcutpoint <= start < cutpoint:
                                newstart = start - prevcutpoint
                                newtag = list(tag)
                                newtag[1] = newstart
                                newtag[2] = (cutpoint - 1
                                             if end > cutpoint
                                             else end - prevcutpoint)
                                linetags[c].append(newtag)
                                if end > cutpoint:
                                    start = cutpoint
                                else:
                                    break
                            prevcutpoint = cutpoint

                    for j, line in enumerate(lines):
                        rowspan = 1 if j < count - 1 else cell.rowspan
                        newcell = Cell(line,  linetags[j],
                                       cell.colspan, cell.rowspan)
                        newrows[j].append(newcell)
                newdata2 += newrows
            else:
                newdata2.append(row)

        data = newdata2

        one_nonempty_col = True
        for row in data:
            nonempty_cols = [cell for cell in row
                             if len(cell.text.strip(' “”"«»„'.decode('utf8'))) > 0]
            if len(nonempty_cols) > 1:
                one_nonempty_col = False
                break

        logging.debug('One empty col?: %s', one_nonempty_col)

        text = tabletext
        tags = [] if tabletags is None else tabletags
        for row in data:
            start = len(text)
            for cell in row:
                offset = len(text)
                text += cell.text
                tags += [self.apply_offset(tag, offset) for tag in cell.tags]
                text += u'\t' if not one_nonempty_col else u''
            text += u'\n'
            end = len(text)
            if not one_nonempty_col:
                tags.append((u'row', start, end))

        globaltabs = [0 for i in range(tabcount)]

        lenmatrix = []

        def zero(globaltabs, i):
            return 0

        for i, row in enumerate(data):
            rowentry = []
            lenmatrix.append(rowentry)
            for cell in row:
                for j in range(cell.colspan):
                    if j == cell.colspan - 1:
                        def lencontrib(text, colspan, globaltabs, n):
                            contrib = len(text) + 1
                            for k in range(colspan - 1):
                                contrib -= globaltabs[n-k-1]
                            return contrib
                        l = functools.partial(lencontrib,
                                              cell.text,
                                              cell.colspan)
                    else:
                        l = zero
                    rowentry.append(l)
            if len(rowentry) < tabcount:
                logging.warn('Bad table, expected %d total column '
                             'span, got only %d', tabcount, len(rowentry))
                while len(rowentry) < tabcount:
                    rowentry.append(zero)

        for i in range(len(globaltabs)):
            cell_lengths = [row[i](globaltabs, i) for row in lenmatrix]
            globaltabs[i] = max(cell_lengths)

        runningsum = 0
        for i, rawtab in enumerate(globaltabs):
            runningsum += rawtab
            globaltabs[i] = runningsum

        if one_nonempty_col:
            tabs = {}
        else:
            tabs = {'': globaltabs}

            for i, row in enumerate(data):
                if any([cell.colspan > 1 for cell in row]):
                    rowtabs = []
                    tabs[i] = rowtabs
                    j=0
                    for cell in row:
                        pos = globaltabs[j+cell.colspan - 1]
                        rowtabs.append(pos)
                        j += cell.colspan

        return text, tags, tabs

    def apply_offset(self, tag, offset):
        mtag = list(tag)
        mtag[1] += offset
        mtag[2] += offset
        return tuple(mtag)

    def txt(self, obj):
        m = "_" + obj.__class__.__name__
        m = getattr(self, m, None)
        if m: # find handler
            return m(obj)
        else:
            logging.debug('No handler for %s, write children', obj)
            return self.process_children(obj)

    def process_children(self, obj, txt=u'', tags=None):
        if tags is None:
            tags = []
        else:
            tags = tags[:]
        for c in obj:
            ctxt, ctags = self.txt(c)
            txtlen = len(txt)
            tags += [self.apply_offset(ctag, txtlen) for ctag in ctags]
            txt += ctxt
        logging.debug('Processed children for %s, returning %s with tags %s',
                      obj, txt, tags)
        return txt, tags

def maketag(name, txt, attrs=None, start=0):
    end = start + len(txt)
    return (name, start, end, attrs) if attrs else (name, start, end)



# -*- coding: utf-8 -*-
# This file is part of Aard Dictionary Tools <http://aarddict.org>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License <http://www.gnu.org/licenses/gpl-3.0.txt>
# for more details.
#
# Copyright (C) 2008-2009  Igor Tkach

from __future__ import with_statement
import functools
import re
import logging
import os
import sys
from itertools import islice

import simplejson

from mwlib import uparser, xhtmlwriter
from mwlib.log import Log
Log.logfile = None

from mwlib import lrucache, expr
expr._cache = lrucache.mt_lrucache(100)

from mwlib.templ.evaluate import Expander
Expander.parsedTemplateCache = lrucache.lrucache(100)

import mwaardwriter

tojson = functools.partial(simplejson.dumps, ensure_ascii=False)

import multiprocessing
from multiprocessing import Pool, TimeoutError
from mwlib.cdbwiki import WikiDB, normname
from mwlib._version import version as mwlib_version
from mwlib.siteinfo import get_siteinfo

import gc

wikidb = None
log = logging.getLogger('wiki')

def _create_wikidb(cdbdir, lang):
    global wikidb
    wikidb = Wiki(cdbdir, lang)

def _init_process(cdbdir, lang):
    global log
    log = multiprocessing.get_logger()
    _create_wikidb(cdbdir, lang)

class ConvertError(Exception):

    def __init__(self, title):
        Exception.__init__(self, title)
        self.title = title

    def __str__(self):
        """
        >>> e = ConvertError('абвгд'.decode('utf8'))
        >>> print str(e)
        ConvertError: абвгд

        """
        return 'ConvertError: %s' % self.title.encode('utf8')

    def __repr__(self):
        """
        >>> e = ConvertError('абвгд'.decode('utf8'))
        >>> print repr(e)
        ConvertError: u'\u0430\u0431\u0432\u0433\u0434'

        """
        return 'ConvertError: %r' % self.title


class EmptyArticleError(ConvertError): pass

def convert(title):
    gc.collect()
    try:
        text = wikidb.getRawArticle(title, resolveRedirect=False)

        if not text:
            raise EmptyArticleError(title)

        redirect = wikidb.get_redirect(text)
        if redirect:
            redirect = normname(redirect)
            meta = {u'r': redirect}
            return title, tojson(('', [], meta)), True

        mwobject = uparser.parseString(title=title,
                                       raw=text,
                                       wikidb=wikidb,
                                       lang=wikidb.lang)
        xhtmlwriter.preprocess(mwobject)
        text, tags = mwaardwriter.convert(mwobject)
    except EmptyArticleError:
        raise
    except Exception:
        log.exception('Failed to process article %s', title.encode('utf8'))
        raise ConvertError(title)
    else:
        return title, tojson((text.rstrip(), tags)), False


class BadRedirect(ConvertError): pass


def parse_redirect(text, aliases):
    """
    >>> aliases = [u"#PATRZ", u"#PRZEKIERUJ", u"#TAM", u"#REDIRECT"]
    >>> parse_redirect(u'#PATRZ [[Zmora]]', aliases)
    u'Zmora'
    >>> parse_redirect(u'#PATRZ[[Zmora]]', aliases)
    u'Zmora'
    >>> parse_redirect(u'#PRZEKIERUJ [[Uwierzytelnianie]]', aliases)
    u'Uwierzytelnianie'

    >>> parse_redirect('#TAM[[Żuraw samochodowy]]'.decode('utf8'), aliases)
    u'\u017buraw samochodowy'

    >>> parse_redirect('#перенапр[[абв]]'.decode('utf8'), [u"#\u043f\u0435\u0440\u0435\u043d\u0430\u043f\u0440"])
    u'\u0430\u0431\u0432'

    >>> parse_redirect('#Перенапр[[абв]]'.decode('utf8'),
    ...                [u"#\u043f\u0435\u0440\u0435\u043d\u0430\u043f\u0440",
    ...                 u"#\u043f\u0435\u0440\u0435\u043d\u0430\u043f\u0440".upper()])
    u'\u0430\u0431\u0432'

    >>> parse_redirect(u'abc', aliases)

    >>> parse_redirect(u'#REDIRECT [[abc', aliases)
    Traceback (most recent call last):
    ...
    BadRedirect: ConvertError: [[abc

    >>> parse_redirect(u'#REDIRECT abc', aliases)
    Traceback (most recent call last):
    ...
    BadRedirect: ConvertError: abc

    >>> parse_redirect('#REDIRECT абв'.decode('utf8'), aliases)
    Traceback (most recent call last):
    ...
    BadRedirect: ConvertError: абв

    """
    for alias in aliases:
        if text.startswith(alias) or text.upper().startswith(alias):
            text = text[len(alias):].lstrip()
            begin = text.find('[[')
            if begin < 0:
                raise BadRedirect(text)
            end = text.find(']]')
            if end < 0:
                raise BadRedirect(text)
            return text[begin+2:end]
    return None

class Wiki(WikiDB):

    def __init__(self, cdbdir, lang):
        WikiDB.__init__(self, cdbdir)
        self.lang = lang
        self.siteinfo = get_siteinfo(self.lang)
        self.redirect_aliases = set()
        aliases = [magicword['aliases']
                                 for magicword in self.siteinfo['magicwords']
                                 if magicword['name'] == 'redirect'][0]

        for alias in aliases:
            self.redirect_aliases.add(alias)
            self.redirect_aliases.add(alias.lower())
            self.redirect_aliases.add(alias.upper())


    def get_siteinfo(self):
        return self.siteinfo

    def get_redirect(self, text):
        return parse_redirect(text, self.redirect_aliases)

    def getTemplate(self, title, followRedirects=True):
        if ":" in title:
            title = title.split(':', 1)[1]
        try:            
            res = self.reader["Template:"+title]
        except KeyError:
            title = normname(title)
            try:
                res = self.reader["Template:"+title]
            except KeyError:
                return ''

        redirect = parse_redirect(res, self.redirect_aliases)
        if redirect:
            redirect = redirect.split("|", 1)[0].split("#", 1)[0]
            if followRedirects:
                return self.getTemplate(redirect, followRedirects=followRedirects)
            else:
                log.warn('Template redirect not followed: %r -> %r' % (title, redirect))
        return res


def total(inputfile, options):
    w = WikiDB(inputfile)
    for i, a in enumerate(islice(w.articles(), options.start, options.end)):
        pass
    try:
        return i+1
    except:
        return 0


def make_input(input_file_name):
    return input_file_name

def collect_articles(input_file, options, compiler):
    p = WikiParser(options, compiler)
    p.parse(input_file)

default_lic_fname = 'license.txt'
default_copyright_fname = 'copyright.txt'
default_metadata_fname = 'metadata.ini'

class WikiParser():

    def __init__(self, options, consumer):
        self.consumer = consumer

        wiki_lang = options.wiki_lang
        metadata_dir = os.path.join(sys.prefix,'share/aardtools/wiki/%s' % wiki_lang)
        default_metadata_dir = os.path.join(sys.prefix,'share/aardtools/wiki/%s' % 'en')

        try:
            siteinfo = get_siteinfo(wiki_lang)
            if siteinfo is None:
                raise Exception('No site info found for %s' % wiki_lang)
        except:
            log.fatal('Failed to read siteinfo for language %(lang)s, '
                      'can\'t proceed. '
                      'Check that siteinfo-%(lang)s.json exists in <MWLIB HOME>/mwlib/siteinfo, '
                      'run "fetch_siteinfo.py %(lang)s" if not', dict(lang=wiki_lang))
            raise SystemExit(1)
        consumer.add_metadata('siteinfo', siteinfo)
        sitename = siteinfo['general']['sitename']
        sitelang = siteinfo['general']['lang']

        metadata_files = []
        if options.metadata:
            metadata_files.append(options.metadata)
        else:
            metadata_files.append(os.path.join(default_metadata_dir, default_metadata_fname))
            metadata_files.append(os.path.join(metadata_dir, default_metadata_fname))

        from ConfigParser import ConfigParser
        c = ConfigParser(defaults={'ver': options.dict_ver,
                                   'lang': wiki_lang,
                                   'update': options.dict_update,
                                   'name': sitename,
                                   'sitelang': sitelang})
        read_metadata_files = c.read(metadata_files)
        if not read_metadata_files:
            log.warn('No metadata files read.')
        else:
            log.info('Using metadata from %s', ', '.join(read_metadata_files))
        for opt in c.options('metadata'):
            value = c.get('metadata', opt)
            self.consumer.add_metadata(opt, value)

        if not options.license and 'license' not in self.consumer.metadata:
            license_file = os.path.join(metadata_dir, default_lic_fname)
            log.info('Looking for license text in %s', license_file)
            if not os.path.exists(license_file):
                log.info('File %s doesn\'t exist', license_file)
                license_file = os.path.join(default_metadata_dir, default_lic_fname)
                log.info('Looking for license text in %s', license_file)
            try:
                with open(license_file) as f:
                    license_text = f.read()
                    self.consumer.add_metadata('license', license_text)
                    log.info('Using license text from %s', license_file)
            except IOError, e:
                log.warn('No license text will be written to the '
                         'output dictionary: %s', str(e))

        if not options.copyright and 'copyright' not in self.consumer.metadata:
            copyright_file = os.path.join(metadata_dir, default_copyright_fname)
            log.info('Looking for copyright notice text in %s', copyright_file)
            if not os.path.exists(copyright_file):
                log.info('File %s doesn\'t exist', copyright_file)
                copyright_file = os.path.join(default_metadata_dir, default_copyright_fname)
                log.info('Looking for copyright notice text in %s', copyright_file)
            try:
                with open(copyright_file) as f:
                    copyright_text = f.read()
                    self.consumer.add_metadata('copyright', copyright_text)
                    log.info('Using copyright notice text from %s', copyright_file)
            except IOError, e:
                log.warn('No copyright notice text will be written to the '
                         'output dictionary: %s', str(e))


        self.lang = wiki_lang
        self.consumer.add_metadata("index_language", sitelang)
        self.consumer.add_metadata("article_language", sitelang)
        log.info('Language: %s (%s)', self.lang, sitelang)

        self.consumer.add_metadata('mwlib',
                                   '.'.join(str(v) for v in mwlib_version))
        self.special_article_re = re.compile(r'^\w+:\S', re.UNICODE)
        self.processes = options.processes if options.processes else None
        self.pool = None
        self.timeout = options.timeout
        self.timedout_count = 0
        self.start = options.start
        self.end = options.end
        if options.nomp:
            log.info('Disabling multiprocessing')
            self.parse = self.parse_simple
        else:
            self.parse = self.parse_mp
        self.mp_chunk_size = options.mp_chunk_size

    def articles(self, f):
        if self.start > 0:
            log.info('Skipping to article %d', self.start)
        _create_wikidb(f, self.lang)
        for title in islice(wikidb.articles(), self.start, self.end):
            if self.special_article_re.match(title):
                self.consumer.skip_article(title)
                continue
            log.debug('Yielding "%s" for processing', title.encode('utf8'))
            yield title
            gc.collect()

    def reset_pool(self, cdbdir, terminate=True):
        if self.pool and terminate:
            log.info('Terminating current worker pool')
            self.pool.terminate()
        log.info('Creating new worker pool with wiki cdb at %s', cdbdir)

        self.pool = Pool(processes=self.processes,
                         initializer=_init_process,
                         initargs=[cdbdir, self.lang])

    def parse_simple(self, f):
        self.consumer.add_metadata('article_format', 'json')
        articles = self.articles(f)
        for a in articles:
            try:
                result = convert(a)
                title, serialized, redirect = result
                self.consumer.add_article(title, serialized, redirect)
            except EmptyArticleError, e:
                self.consumer.empty_article(e.title)
            except ConvertError, e:
                self.consumer.fail_article(e.title)

    def parse_mp(self, f):
        try:
            self.consumer.add_metadata('article_format', 'json')
            articles = self.articles(f)
            self.reset_pool(f)
            iter_count = 1
            while True:
                if iter_count:
                    chunk = islice(articles, self.mp_chunk_size)
                    iter_count = 0
                else:
                    break

                resulti = self.pool.imap_unordered(convert, chunk)
                while True:
                    try:
                        result = resulti.next(self.timeout)
                        iter_count += 1
                        title, serialized, redirect = result
                        self.consumer.add_article(title, serialized,redirect)
                    except StopIteration:
                        break
                    except TimeoutError:
                        log.warn('Worker pool timed out')
                        self.consumer.timedout(count=len(multiprocessing.active_children()))
                        self.reset_pool(f)
                        resulti = self.pool.imap_unordered(convert, chunk)
                    except AssertionError:
                        log.exception()
                    except EmptyArticleError, e:
                        self.consumer.empty_article(e.title)
                    except ConvertError, e:
                        self.consumer.fail_article(e.title)
                    except KeyboardInterrupt:
                        log.error('Keyboard interrupt: '
                                  'terminating worker pool')
                        self.pool.terminate()
                        raise

                self.pool.close()
                self.reset_pool(f, terminate=False)
        finally:
            self.pool.close()
            self.pool.join()



import os
import sys
from setuptools import setup, find_packages

import aardtools

wiki_meta_en = ['wiki/en/copyright.txt',
                'wiki/en/license.txt',
                'wiki/en/metadata.ini']

setup(
    name = aardtools.__name__,
    version = aardtools.__version__,
    packages = find_packages(),
    entry_points = {
        'console_scripts': ['aardcompile = aardtools.compiler:main',
                            'aardc = aardtools.compiler:main']
    },

    install_requires = [ 'aarddict == 0.7.5',
                         'PyICU >= 0.8.1',
                         'mwlib == 0.11.2',
                         'simplejson',
                         'multiprocessing'],

    data_files = [
        (os.path.join(sys.prefix,'share/aardtools/wiki/en'), wiki_meta_en)
        ],

    author = "Igor Tkach",
    author_email = "itkach@aarddict.org",
    description =  '''Tools to create dictionaries in aarddict format.''',
    license = "GPL 3",
    keywords = ['aarddict', 'aardtools', 'wiki', 'wikipedia',
                'xdxf', 'dict', 'dictionary', 'maemo'],
    url = "http://aarddict.org",
    classifiers=[
                 'Development Status :: 3 - Alpha',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python',
                 'License :: OSI Approved :: GNU General Public License (GPL)',
                 'Topic :: Utilities',
                 'Environment :: Console'
    ]
)




#!/usr/bin/python
"""
sortexternal.py:  Sort files larger than available memory

There are no restrictions on record contents (e.g. record may contain \n or \x00).
Records do not need to end with \n or \x00.
Records may be up to 2^31 bytes long.

Uses ideas from:
  Title: Sorting big files the Python 2.4 way
  Submitter: Nicolas Lehuen
  http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/466302

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2008  Jeremy Mortis

"""

from heapq import heappop, heappush
import os
import tempfile
import struct
from functools import partial

class VariableLengthRecordFile(file):

    def __init__(self, name, mode, bufsize=-1):
        file.__init__(self, name, mode, bufsize)
        self.headerFormat = "i"
        self.headerLength = struct.calcsize(self.headerFormat)
        self._pack = partial(struct.pack, self.headerFormat)

    def readline(self):
        header = self.read(self.headerLength)
        if header == "":
            return (-2, "")

        recordLength = struct.unpack(self.headerFormat, header)[0]
        if recordLength == -1:
            return (-1, "")

        return (1, self.read(recordLength))

    def writeline(self, s):
        self.write(self._pack(len(s)))
        self.write(s)

    def mark(self):
        self.write(self._pack(-1))

class SortExternal:

    def __init__(self, buffer_size=200000, filenum=16, work_dir=None):
        self.buffer_size = buffer_size
        if work_dir:
            if not os.path.exists(work_dir):
                os.mkdir(work_dir)
            self.work_dir = work_dir
        else:
            self.work_dir = tempfile.mkdtemp()
        self.chunk = []
        self.chunksize = 0

        self.inputChunkFiles = []
        self.outputChunkFiles = []

        for i in range(filenum):
            filename = os.path.join(self.work_dir, "sort-%06i" % i)
            self.inputChunkFiles.append(VariableLengthRecordFile(filename,
                                                                 'w+b',
                                                                 8*1024))
        for i in range(filenum, filenum * 2):
            filename = os.path.join(self.work_dir, "sort-%06i" %i )
            self.outputChunkFiles.append(VariableLengthRecordFile(filename,
                                                                  'w+b',
                                                                  8*1024))

        self.currOutputFile = -1
        self.chunkDepth = 1


    def __iter__(self):
        return self

    def put(self, value):

        self.chunk.append(value)
        self.chunksize += len(value)

        if self.chunksize < self.buffer_size:
            return

        self.chunk.sort()
        self.put_chunk(self.chunk)
        self.chunk = []
        self.chunksize = 0

    def put_chunk(self, valueIterator):

        self.currOutputFile += 1
        if self.currOutputFile >= len(self.outputChunkFiles):
            self.currOutputFile = 0
            self.chunkDepth += 1

        out_file = self.outputChunkFiles[self.currOutputFile]
        for value in valueIterator:
            out_file.writeline(value)
        out_file.mark()

    def sort(self):

        if len(self.chunk) > 0:
            self.chunk.sort()
            self.put_chunk(self.chunk)

        while self.chunkDepth > 1:
            self.mergeFiles()

        t = self.inputChunkFiles
        self.inputChunkFiles = self.outputChunkFiles
        self.outputChunkFiles = t

        for f in self.inputChunkFiles:
            f.flush()
            f.seek(0)

        self.prepareChunkMerge()

    def prepareChunkMerge(self):

        self.chunkHeap = []

        for chunkFile in self.inputChunkFiles:
            status, value = chunkFile.readline()
            if status > 0:
                heappush(self.chunkHeap,(value,chunkFile))

    def mergeFiles(self):

        t = self.inputChunkFiles
        self.inputChunkFiles = self.outputChunkFiles
        self.outputChunkFiles = t

        self.currOutputFile = -1
        self.chunkDepth = 1

        for f in self.outputChunkFiles:
            f.flush()
            f.truncate(0)
            f.seek(0)

        for f in self.inputChunkFiles:
            f.flush()
            f.seek(0)

        # for each layer of chunks
        while True:
            self.prepareChunkMerge()
            if not self.chunkHeap:
                break
            self.put_chunk(self)

    def next(self):
        # merges current chunk layer
        if not self.chunkHeap:
            raise StopIteration

        value, chunkFile = heappop(self.chunkHeap)

        returnValue = value
        status, value = chunkFile.readline()
        if status > 0:
            heappush(self.chunkHeap, (value, chunkFile))

        return returnValue


    def cleanup(self):

        for chunkFile in self.inputChunkFiles:
            chunkFile.close()
            os.remove(chunkFile.name)

        for chunkFile in self.outputChunkFiles:
            chunkFile.close()
            os.remove(chunkFile.name)

        os.rmdir(self.work_dir)
